#include "Arduino.h"
#include "SimpleSim800.h"
#include <DebugOverSoftSerial.h>

//This enables debugging mode, to disable it - set value to 0
#define DEBUG 1 

DebugOverSoftSerial gsmDebugLogger;

// alte congstar-Karten
/*
const char _networkAPN[] = "internet.t-mobile";
const char _userName[] = "t-mobile";
const char _passWord[] = "tm";
*/

// neue congstar-Karten
/*
const char _networkAPN[] = "internet.telekom";
const char _userName[] = "congstar";
const char _passWord[] = "cs";
*/

// 1NCE
const char _networkAPN[] = "iot.1nce.net";
const char _userName[] = "";
const char _passWord[] = "";

const char _simPin[] = "9545";
const char _dnsServer[] = "";

// Initialize the constructors
SimpleSim800::SimpleSim800()
{
	_isModuleConfigured = false;
	_isConnected = false;
}
 
void SimpleSim800::begin()
{
	Serial.begin(9600);
	while (_attention(3) != 0);
    _attentionDisableEcho();
}

int SimpleSim800::init()
{
	if (_attention(3) == COMMAND_ERROR)
	{
		return INIT_AT_FAILED;
	}
    return INIT_SUCCESS;
}

int SimpleSim800::connect()
{
	if (_isConnected)
	{
		return CONNECT_SUCCESS;
	}
	
	if (_attentionEnablePhoneFunctionality() == COMMAND_ERROR)
	{
		return INIT_ENABLE_PHONE_FUNCTIONALITY_FAILED;
	}
	if (_attentionCheckAndSubmitPin() == COMMAND_ERROR)
	{
		return INIT_SUBMIT_PIN_FAILED;
	}
	
	_attentionSignalQuality();
	// _attentionListOperators();
	_attentionRegistrationState();
	
	int resultConnectNetwork = _attentionConnectToNetwork();
	if (resultConnectNetwork == COMMAND_ERROR)
	{
		return CONNECT_NETWORK_FAILED;
	}
	
	int resultEnableBearer = _attentionEnableBearer();
	if (resultEnableBearer == COMMAND_ERROR)
	{
		return CONNECT_ENABLE_BEARER_FAILED;
	}
	
	int resultConfigureNtp = _attentionConfigureNtp();
	if (resultConfigureNtp == COMMAND_ERROR)
	{
		return ENABLE_NTP_FAILED;
	}

	_isModuleConfigured = true;
	_isConnected = true;
	return CONNECT_SUCCESS;
}

int SimpleSim800::disconnect()
{
	_isConnected = false;
	
    if (_attentionDisonnectToNetwork() == COMMAND_ERROR)
	{
		return DISCONNECT_NETWORK_FAILED;
	}
	
    if (_attentionDisablePhoneFunctionality() == COMMAND_ERROR)
	{
		return INIT_DISABLE_PHONE_FUNCTIONALITY_FAILED;
	}
	return COMMAND_SUCCESS;
}

int SimpleSim800::getDateTime(char* dateTimeString)
{
	_println(F("AT+CCLK?"));

	unsigned long timerStart, timerEnd;
    timerStart = millis();

	bool concatToResult = false;

    while (1) {
        if (Serial.available()) 
		{
            char c = Serial.read();
			char s[2] = {0};
			#if DEBUG
			gsmDebugLogger.print(String(c));
			#endif
			if (concatToResult)
			{
                if (c == '"')
				{
                    break;
				}
				s[0] = c;
				strcat(dateTimeString, s);
			}
            else if (c == '"')
			{
				concatToResult = true;
			}
        }
        timerEnd = millis();
        if (timerEnd - timerStart > 1000) {
            return -1;
        }
    }
    _readRemainingCharacters();
	
    return GET_TIME_SUCCESS;
}

int SimpleSim800::getImei(char* str)
{
	_println(F("AT+GSN"));
	unsigned long timerStart, timerEnd;
    timerStart = millis();
	bool readingImei = false;
	while(1)
	{
        if (Serial.available()) 
		{
            char c = Serial.read();
			char s[2] = {0};
			#if DEBUG
			gsmDebugLogger.print(String(c));
			#endif
			if (c == '0' || c == '1' || c == '2' || c == '3' || c == '4' || 
			    c == '5' || c == '6' || c == '7' || c == '8' || c == '9')
			{
				readingImei = true;
				s[0] = c;
				strcat(str, s);
			}
            if (readingImei && (c == '\r' || c == '\n'))
			{
				break;
			}
		}
		timerEnd = millis();
        if (timerEnd - timerStart > 1000) {
            return -1;
        }
	}
	_readRemainingCharacters();
	return GET_IMEI_SUCCESS;
}

int SimpleSim800::postData(const char* url, const char* data)
{
	if (_sendCmdAndWaitForResp(F("AT+HTTPINIT"), "OK", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	if (_sendCmdAndWaitForResp(F("AT+HTTPPARA=CID,1"), "OK", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	_print(F("AT+HTTPPARA=\"URL\",\""));
	_print(url);
	if (_sendCmdAndWaitForResp(F("\""), "OK", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	_print(F("AT+HTTPPARA=\"CONTENT\",\""));
	_print(F("application/x-www-form-urlencoded"));
	if (_sendCmdAndWaitForResp(F("\""), "OK", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	int dataLength = strlen(data);
	char cstr[16];
    itoa(dataLength, cstr, 10);
	_print(F("AT+HTTPDATA="));
	_print(cstr);
	if (_sendCmdAndWaitForResp(F(",1000"), "DOWNLOAD", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	_print(data);
	delay(1100);
	if (_sendCmdAndWaitForResp(F("AT+HTTPACTION=1"), "+HTTPACTION: 1,200", DEFAULT_TIMEOUT * 5) != COMMAND_SUCCESS)
	{
		if (_sendCmdAndWaitForResp(F("AT+HTTPTERM"), "OK", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
		{
			return POST_DATA_FAILED;
		}
		return POST_DATA_FAILED;
	}
	_readRemainingCharacters();
	
	delay(1000);
	_println(F("AT+HTTPREAD=0,20"));
	for(int i = 0; i < 100; i++)
	{
		delay(600);
		if (Serial.available())
		{
			break;
		}
	}
	while (Serial.available()) {
		char c = Serial.read();
		#if DEBUG
		gsmDebugLogger.print(String(c));
		#endif
	}

	/*
	if (_sendCmdAndWaitForResp(F("AT+HTTPREAD"), "OK", DEFAULT_TIMEOUT * 10) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	*/
	
	if (_sendCmdAndWaitForResp(F("AT+HTTPTERM"), "OK", DEFAULT_TIMEOUT) != COMMAND_SUCCESS)
	{
		return POST_DATA_FAILED;
	}
	return COMMAND_SUCCESS;
}


int SimpleSim800::enablePowerSaving()
{
	// return COMMAND_SUCCESS;
	return _sendCmdAndWaitForResp(F("AT+CSCLK=2"), "OK", DEFAULT_TIMEOUT);
}

int SimpleSim800::disablePowerSaving()
{
	// return COMMAND_SUCCESS;
	int result = 0;
	int counter = 0;
	do
	{
		result = _sendCmdAndWaitForResp(F("AT+CSCLK=0"), "OK", DEFAULT_TIMEOUT);
		counter++;
	}
	while (result != COMMAND_SUCCESS && counter < 20);
	return result;
}


int SimpleSim800::_readBuffer(char* buffer, int count, unsigned int timeOut) 
{
    int i = 0;
    unsigned long timerStart, timerEnd;
    timerStart = millis();
    while (1) {
        while (Serial.available()) {
            char c = Serial.read();
			/*
            if (c == '\r' || c == '\n') {
                c = '$';
            }
			*/
            buffer[i++] = c;
            if (i > count - 1) {
                break;
            }
        }
        if (i > count - 1) {
            break;
        }
        timerEnd = millis();
        if (timerEnd - timerStart > 1000 * timeOut) {
            break;
        }
    }
    #if DEBUG
	gsmDebugLogger.print(buffer);
	#endif
    delay(500);
    while (Serial.available()) {  // display the other thing..
        char c = Serial.read();
		#if DEBUG
		gsmDebugLogger.print(String(c));
		#endif
    }
	
	
    return 0;
}

void SimpleSim800::_cleanBuffer(char* buffer, int count) 
{
    for (int i = 0; i < count; i++) {
        buffer[i] = '\0';
    }
}

int SimpleSim800::_waitForResp(const char* resp, unsigned int timeout) 
{
    int len = strlen(resp);
    int sum = 0;
    unsigned long timerStart, timerEnd;
    timerStart = millis();

    while (1) {
        if (Serial.available()) {
            char c = Serial.read();
			#if DEBUG
			gsmDebugLogger.print(String(c));
			#endif
            sum = (c == resp[sum]) ? sum + 1 : 0;
            if (sum == len) {
                break;
            }
        }
        timerEnd = millis();
        if (timerEnd - timerStart > 1000 * timeout) {
            return -1;
        }
    }
    _readRemainingCharacters();

    return 0;
}

void SimpleSim800::_sendEndMark(void) 
{
    _print(String((char)26));
}

int SimpleSim800::_sendCmdAndWaitForResp(String cmd, const char* resp, unsigned timeout) 
{
    _println(cmd);
    return _waitForResp(resp, timeout);
}

uint32_t SimpleSim800::_str_to_ip(const char* str) 
{
    uint32_t ip = 0;
    char* p = (char*)str;

    for (int i = 0; i < 4; i++) 
	{
        ip |= atoi(p);
        p = strchr(p, '.');
        if (p == NULL) 
		{
            break;
        }
        if (i < 3) 
		{
            ip <<= 8;
        }
        p++;
    }
    return ip;
}

void SimpleSim800::_print(String cmd)
{
	Serial.print(cmd);
	#if DEBUG
	gsmDebugLogger.print(cmd);
	#endif
}

void SimpleSim800::_println(String cmd)
{
	Serial.println(cmd);
	#if DEBUG
	gsmDebugLogger.println(cmd);
	#endif
}

int SimpleSim800::_sendCmdAndWaitForResp(const char* cmd, const char* resp, unsigned timeout) 
{
    _println(cmd);
    return _waitForResp(resp, timeout);
}

void SimpleSim800::_print(const char* cmd)
{
	Serial.print(cmd);
	#if DEBUG
	gsmDebugLogger.print(cmd);
	#endif
}

void SimpleSim800::_println(const char* cmd)
{
	Serial.println(cmd);
	#if DEBUG
	gsmDebugLogger.println(cmd);
	#endif
}

void SimpleSim800::_readRemainingCharacters()
{
	while (Serial.available()) {
        char c = Serial.read();
		#if DEBUG
		gsmDebugLogger.print(String(c));
		#endif
    }
}

int SimpleSim800::_attention() 
{
    return _attention(1);
}

int SimpleSim800::_attention(int timeoutFactor) 
{
    int ret = _sendCmdAndWaitForResp("AT", "OK", DEFAULT_TIMEOUT * timeoutFactor);
    return ret;
}

int SimpleSim800::_attentionDisableEcho()
{
    if (_sendCmdAndWaitForResp(F("ATE0"), "OK", DEFAULT_TIMEOUT) == COMMAND_ERROR) 
	{
		return COMMAND_ERROR;
    }	
	return COMMAND_SUCCESS;
}

int SimpleSim800::_attentionEnablePhoneFunctionality()
{
	int result = disablePowerSaving();
    if (result != COMMAND_SUCCESS)
	{
		return result;
	}
		
	// AT+CFUN Max Response Time 10 Seconds
    if (_sendCmdAndWaitForResp(F("AT+CFUN=1"), "OK", DEFAULT_TIMEOUT * 2) == COMMAND_ERROR) 
	{
        return COMMAND_ERROR;
    }
	return COMMAND_SUCCESS;
}

int SimpleSim800::_attentionDisablePhoneFunctionality()
{
	// AT+CFUN Max Response Time 10 Seconds
    if (_sendCmdAndWaitForResp(F("AT+CFUN=0"), "OK", DEFAULT_TIMEOUT * 2) == COMMAND_ERROR) 
	{
        return COMMAND_ERROR;
    }
	if (enablePowerSaving() == COMMAND_ERROR) 
	{
        return COMMAND_ERROR;
    }
	return COMMAND_SUCCESS;
}

int SimpleSim800::_attentionCheckAndSubmitPin()
{
	// AT+CPIN Max Response Time 5 Seconds
	if (_sendCmdAndWaitForResp(F("AT+CPIN?"), "+CPIN: SIM PIN", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
	{
		if (strlen(_simPin) > 0)
		{
			_print(F("AT+CPIN="));
			if (_sendCmdAndWaitForResp(_simPin, "OK", DEFAULT_TIMEOUT * 2) == COMMAND_SUCCESS)
			{
				delay(5000);
				_readRemainingCharacters();
				return COMMAND_SUCCESS;
			}
			else
			{
				_readRemainingCharacters();
				return COMMAND_ERROR;
			}
		}
		else
		{
			return COMMAND_ERROR;
		}
		
	}
	else if (_sendCmdAndWaitForResp(F("AT+CPIN?"), "+CPIN: READY", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
	{
		return COMMAND_SUCCESS;
	}
	return COMMAND_ERROR;
}

int SimpleSim800::_attentionSignalQuality()
{
    if (_sendCmdAndWaitForResp(F("AT+CSQ"), "+CSQ", DEFAULT_TIMEOUT * 2) == COMMAND_SUCCESS) 
	{
		delay(1000);
		_readRemainingCharacters();
        return COMMAND_SUCCESS;
    }
	return COMMAND_ERROR;
}

int SimpleSim800::_attentionListOperators()
{
    if (_sendCmdAndWaitForResp(F("AT+COPS=?"), "+COPS", DEFAULT_TIMEOUT * 10) == COMMAND_SUCCESS) 
	{
		delay(2000);
		_readRemainingCharacters();
        return COMMAND_SUCCESS;
    }
	return COMMAND_ERROR;
}

int SimpleSim800::_attentionConnectToNetwork()
{
	// AT+CGATT Max Response Time 10 Seconds
    if (_sendCmdAndWaitForResp(F("AT+CGATT?"), "+CGATT: 1", DEFAULT_TIMEOUT * 2) == COMMAND_ERROR) 
	{
		int result = -1;
		byte counter = 0;
		do
		{
			result = _sendCmdAndWaitForResp(F("AT+CGATT=1"), "OK", DEFAULT_TIMEOUT * 5);
			counter++;
		}
		while(result != COMMAND_SUCCESS && counter < 3);
		return result;
    }
	return COMMAND_SUCCESS;
}

int SimpleSim800::_attentionDisonnectToNetwork()
{
	_sendCmdAndWaitForResp(F("AT+SAPBR=0,1"), "OK", DEFAULT_TIMEOUT * 17);
	
	// AT+CGATT Max Response Time 10 Seconds
    if (_sendCmdAndWaitForResp(F("AT+CGATT=0"), "OK", DEFAULT_TIMEOUT * 2) == COMMAND_ERROR) 
	{
        return COMMAND_ERROR;
    }
	return COMMAND_SUCCESS;
}

int SimpleSim800::_attentionRegistrationState()
{
	// AT+CREG? Max Response Time 10 Seconds
	int counter = 0;
	while(counter < 10)
	{
		if (_sendCmdAndWaitForResp(F("AT+CREG?"), "+CREG: 0,1", DEFAULT_TIMEOUT * 2) == COMMAND_SUCCESS) 
		{
			break;
		}
		if (_sendCmdAndWaitForResp(F("AT+CREG?"), "+CREG: 0,5", DEFAULT_TIMEOUT * 2) == COMMAND_SUCCESS) 
		{
			break;
		}
		delay(3000);
		counter++;
	}
	return COMMAND_SUCCESS;
}

int SimpleSim800::_attentionEnableBearer()
{
	if (_isModuleConfigured == false)
	{
		// Configure bearer profile 1
		// AT+SAPBR Max Response Time for open bearer 85 Seconds
		if (_sendCmdAndWaitForResp(F("AT+SAPBR=3,1,\"Contype\",\"GPRS\""), "OK", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
		{
			_print(F("AT+SAPBR=3,1,\"APN\",\""));
			_print(_networkAPN);
			if (_sendCmdAndWaitForResp(F("\""), "OK", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
			{
				if (strlen(_userName) > 0)
				{
					_print(F("AT+SAPBR=3,1,\"USER\",\""));
					_print(_userName);
					_sendCmdAndWaitForResp(F("\""), "OK", DEFAULT_TIMEOUT);
					_print(F("AT+SAPBR=3,1,\"PWD\",\""));
					_print(_passWord);
					_sendCmdAndWaitForResp(F("\""), "OK", DEFAULT_TIMEOUT);
				}
				delay(2000);
			}
			else
			{
				return COMMAND_ERROR;
			}
		}
		else
		{
			return COMMAND_ERROR;
		}
	}
	
	// Open a GPRS context
	if (_sendCmdAndWaitForResp(F("AT+SAPBR=1,1"), "OK", DEFAULT_TIMEOUT * 5) == COMMAND_SUCCESS)
	{
		delay(2000);
		_readRemainingCharacters();
		// Query a GPRS context
		if (_sendCmdAndWaitForResp(F("AT+SAPBR=2,1"), "OK", DEFAULT_TIMEOUT * 15) == COMMAND_SUCCESS)
		{
			delay(2000);
			_readRemainingCharacters();
			return COMMAND_SUCCESS;
		}
		else
		{
			return COMMAND_ERROR;
		}
	}
	else
	{
		_readRemainingCharacters();
		return COMMAND_ERROR;
	}
    return COMMAND_ERROR;
}

int SimpleSim800::_attentionConfigureNtp()
{
	if (_isModuleConfigured == true)
	{
		return COMMAND_SUCCESS;
	}
	if (_sendCmdAndWaitForResp(F("AT+CNTPCID=1"), "OK", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
	{
		if (_sendCmdAndWaitForResp(F("AT+CNTP=\"pool.ntp.org\",0"), "OK", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
		{
			if (_sendCmdAndWaitForResp(F("AT+CNTP"), "OK", DEFAULT_TIMEOUT) == COMMAND_SUCCESS)
			{
				delay(5000);
				bool isClockOk = false;
				do
				{
					_println(F("AT+CCLK?"));
					bool isStartReading = false;
					int characterCounter = 0;
					char yearBuffer[3] = {0};
					unsigned long timerStart, timerEnd;
					timerStart = millis();
					while(1)
					{
						if (Serial.available()) 
						{
							char c = Serial.read();
							#if DEBUG
							gsmDebugLogger.print(String(c));
							#endif
							if(isStartReading)
							{
								if (c == '0' || c == '1' || c == '2' || c == '3' || c == '4' || 
									c == '5' || c == '6' || c == '7' || c == '8' || c == '9')
								{
									yearBuffer[characterCounter] = c;
									characterCounter++;
								}
							}
							if (c == '0')
							{
								isStartReading = true;
							}
							if (characterCounter = 2)
							{
								break;
							}
						}
						timerEnd = millis();
						if (timerEnd - timerStart > 1000) 
						{
							_readRemainingCharacters();
							return COMMAND_ERROR;
						}
					}
					_readRemainingCharacters();
					
					// read two characters from date
					int year;
					sscanf(yearBuffer, "%d", &year);
					if (year != 4)
					{
						isClockOk = true;
					}
					delay(2000);
				}
				while(!isClockOk);
        		return COMMAND_SUCCESS;
			}
			// hier ntp wieder abschalten?
			return COMMAND_ERROR;
		}
		return COMMAND_ERROR;
	}	
	return COMMAND_ERROR;
}

boolean SimpleSim800::isSummertime(int year, byte month, byte day, byte hour, byte tzHours)
// European Daylight Savings Time calculation by "jurs" for German Arduino Forum
// input parameters: "normal time" for year, month, day, hour and tzHours (0=UTC, 1=MEZ)
// return value: returns true during Daylight Saving Time, false otherwise
{
  if (month<3 || month>10) 
  {
    return false; // keine Sommerzeit in Jan, Feb, Nov, Dez
  }
  if (month>3 && month<10) 
  {
	return true; // Sommerzeit in Apr, Mai, Jun, Jul, Aug, Sep
  }
  if (month==3 && (hour + 24 * day)>=(1 + tzHours + 24*(31 - (5 * year /4 + 4) % 7)) || month==10 && (hour + 24 * day)<(1 + tzHours + 24*(31 - (5 * year /4 + 1) % 7)))
  {
    return true;
  }
  return false;
}