#ifndef SimpleSim800_h
#define SimpleSim800_h

#define DEFAULT_TIMEOUT     5

#define COMMAND_SUCCESS  0
#define COMMAND_ERROR   -1

#define INIT_SUCCESS 							    0
#define INIT_AT_FAILED						     -100
#define INIT_ENABLE_PHONE_FUNCTIONALITY_FAILED   -101
#define INIT_SUBMIT_PIN_FAILED                   -102

#define CONNECT_SUCCESS								0
#define CONNECT_NETWORK_FAILED					 -200
#define CONNECT_ENABLE_BEARER_FAILED			 -201

#define ENABLE_NTP_FAILED                        -300

#define GET_TIME_SUCCESS                            0
#define GET_IMEI_SUCCESS                            0
#define POST_DATA_SUCCESS                           0
#define POST_DATA_FAILED                         -400

#define DISCONNECT_NETWORK_FAILED                -500
#define INIT_DISABLE_PHONE_FUNCTIONALITY_FAILED  -600

class SimpleSim800
{
	private:
	
	// member variables
	bool _isModuleConfigured;
	bool _isConnected;
	
	uint32_t _ip;
	
	// basic mathods for communication
	int _readBuffer(char* buffer, int count, unsigned int timeOut = DEFAULT_TIMEOUT);
	void _cleanBuffer(char* buffer, int count);
	int _waitForResp(const char* resp, unsigned timeout);
	void _sendEndMark(void);
	int _sendCmdAndWaitForResp(const char* cmd, const char* resp, unsigned timeout);
	void _print(const char* cmd);
	void _println(const char* cmd);
	int _sendCmdAndWaitForResp(String cmd, const char* resp, unsigned timeout);
	void _print(String cmd);
	void _println(String cmd);
	void _readRemainingCharacters();

    // helper methods
    uint32_t _str_to_ip(const char* str);

    // methods für basic settings
    int _attention();
	int _attention(int timeoutFactor);
	int _attentionDisableEcho();
	int _attentionEnablePhoneFunctionality();
	int _attentionDisablePhoneFunctionality();
	int _attentionCheckAndSubmitPin();
	int _attentionSignalQuality();
	int _attentionListOperators();
	int _attentionRegistrationState();
	int _attentionConnectToNetwork();
	int _attentionDisonnectToNetwork();
	int _attentionEnableBearer();
	int _attentionConfigureNtp();
	
	public:
	SimpleSim800();
	
	void begin();
	int init();
	int connect();
	int disconnect();
	int enablePowerSaving();
	int disablePowerSaving();

	int getDateTime(char* str);
	bool isSummertime(int year, byte month, byte day, byte hour, byte tzHours);
	
	int getImei(char* str);
	
	int postData(const char* url, const char* data);
};

 #endif